/*
 * PIC32 microcontroller initialisation code.
 *
 * Copyright (C) 2017, 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "mips.h"
#include "pic32.h"

/* Application-specific configuration. */

#include "devconfig.h"

/* Disable JTAG functionality on pins. */

.section .devcfg0, "a"
.word 0xfffffffb	/* DEVCFG0<2> = JTAGEN = 0 */

/* Select oscillator, pin usage, watchdog timer and peripheral clock divider. */

.section .devcfg1, "a"
.word (DEVCFG1_UNUSED | DEVCFG1_CONFIG)

/* Set clock dividers and multiplier. */

.section .devcfg2, "a"
.word (DEVCFG2_UNUSED | DEVCFG2_CONFIG)

/* The start routine is placed at the boot location. */

.section .boot, "a"

.globl _start
.extern relocate_data
.extern _data_original
.extern _data
.extern _data_size
.extern main

_start:
	/* Enable caching. */

	mfc0 $v1, CP0_CONFIG
	li $t8, ~CONFIG_K0
	and $v1, $v1, $t8
	ori $v1, $v1, CONFIG_K0_CACHABLE_NONCOHERENT
	mtc0 $v1, CP0_CONFIG
	nop

	/* Get the RAM size. */

	la $v1, BMXDRMSZ
	lw $t0, 0($v1)

        /* Initialise the stack pointer. */

        li $v1, KSEG0_BASE
        addu $sp, $t0, $v1      /* sp = KSEG0_BASE + RAM size */

        /* Initialise the globals pointer. */

        lui $gp, %hi(_GLOBAL_OFFSET_TABLE_)
        ori $gp, $gp, %lo(_GLOBAL_OFFSET_TABLE_)

	/*
	Relocate mutable data and then jump to the main program.

	Since the boot code is separate from the other code, addresses cannot be
	obtained via the GOT:

	"relocation truncated to fit: R_MIPS_PC16 against `main'"
	*/

	lui $a0, %hi(_data_original)
	ori $a0, $a0, %lo(_data_original)

	lui $a1, %hi(_data)
	ori $a1, $a1, %lo(_data)

	lui $a2, %hi(_data_size)
	ori $a2, $a2, %lo(_data_size)
	lw $a2, 0($a2)

	lui $t9, %hi(relocate_data)
	ori $t9, $t9, %lo(relocate_data)
	jal $t9
	nop

        lui $t9, %hi(main)
        ori $t9, $t9, %lo(main)
	jr $t9
	nop
